#include <stdio.h>
#include <stdlib.h>
#include <time.h>

typedef struct _point {
    double x;
    double y;
} Point;

/*
 * Γράψτε έναν constructor που δημιουργεί και επιστρέφει έναν πίνακα από N
 * τέτοια «σημεία» εντός ενός ορθογωνίου παραλληλογράμμου διαστάσεων a x b
 * το οποίο ξεκινά από την αρχή των αξόνων. Ως constructor, εάν κάτι δεν
 * είναι λογικό, επιστρέφει NULL.
 */
Point *ptaCreateRandomInBox(double a, double b, int N) {
    /*
     * Πάντα ελέγχουμε τα δεδομένα εάν είναι "λογικά" ή εντός προδιαγραφών
     */
    if (N < 1) {
        return NULL;
    }

    /*
     * Ζητάμε την απαιτούμενη μνήμη Ν φορές το μέγεθος του struct
     */
    Point *points = (Point *) malloc(N * sizeof(Point));
    /*
     * Πάντα μετά τη malloc το πρώτο πράγμα που κάνουμε είναι να
     * ελέγχουμε εάν είναι NULL (εάν απέτυχε δηλαδή να βρεθεί η
     * απαιτούμενη μνήμη)
     */
    if (points == NULL) {
        /*
         * 1. Πρώτα εξετάζουμε εάν μπορεί/έχει νόημα να συνεχίσει
         * ο κώδικας χωρίς τη μνήμη αυτή
         * 2. Εάν πρέπει να τερματιστεί ο κώδικας φροντίζουμε να
         * αποδεσμεύσουμε τη μνήμη ή οποία άλλα resources έχουμε
         * τυχόν χρησιμοποιήσει και δεν χρειάζονται πλέον
         */
        return NULL;
    }

    /*
     * Για κάθε σημείο από αυτά, θέτω τις συντεταγμένες σε μία
     * τυχαία τιμή από το 0 μέχρι το a ή το b.
     */
    for (int i = 0; i < N; ++i) {
        /*
         * ΠΡΟΣΟΧΗ! Η σειρά στο παρακάτω "γινόμενο" είναι σημαντική
         * καθώς εάν το a δεν είναι πρώτος όρος, ώστε να μετατρέπει
         * τα ενδιάμεσα αποτελέσματα σε double, τότε το rand()/RAND_MAX
         * είναι πηλίκο ακεραίων και επειδή το RAND_MAX είναι το όριο
         * (δηλδαή μεγαλύτερο από) της rand(), προκύπτει πάντα 0.
         */
        points[i].x = a * rand() / RAND_MAX;
        points[i].y = b * rand() / RAND_MAX;
    }

    /*
     * Ο πίνακας των "σημείων" είναι έτοιμος και τον επιστρέφουμε
     */
    return points;
}

/*
 * Γράψτε μία συνάρτηση που επεκτείνει έναν τέτοιο πίνακα σημείων,
 * προσθέτοντάς τους άλλα Μ σημεία εντός του a x b, όχι απαραίτητα
 * ίδιου με το αρχικό. Αυτή θα επιστρέφει έναν νέο δείκτη στον πίνακα
 * ή NULL, εάν δεν γίνεται να επεκταθεί.
 */
Point *ptaAppendRandomInBox(double a, double b, int M, Point *pts, int N) {
    /*
     * Πάντα ελέγχω τα δεδομένα
     */
    if (M < 1 || N < 1 || pts == NULL) {
        return NULL;
    }

    /*
     * Χρησιμοποιώ την realloc, ώστε να παραμείνουν τα υπάρχοντα στοιχεία
     * και να χρειαστεί να προσθέσω μόνο τα επιπλέον (από Ν έως Ν+Μ)
     */
    Point *points = (Point *) realloc(pts, (M + N) * sizeof(Point));
    /*
     * Πάντα ελέγχω το αποτέλεσμα των malloc/realloc
     */
    if (points == NULL) {
        /*
         * 1. Πρώτα εξετάζουμε εάν μπορεί/έχει νόημα να συνεχίσει
         * ο κώδικας χωρίς τη μνήμη αυτή
         * 2. Εάν πρέπει να τερματιστεί ο κώδικας φροντίζουμε να
         * αποδεσμεύσουμε τη μνήμη ή οποία άλλα resources έχουμε
         * τυχόν χρησιμοποιήσει και δεν χρειάζονται πλέον
         */
        return NULL;
    }

    /*
     * Για κάθε σημείο από το Ν (τα προηγούμενα ήταν 0 ως Ν-1) μέχρι και το Ν+Μ-1
     */
    for (int i = N; i < N + M; ++i) {
        points[i].x = a * rand() / RAND_MAX;
        points[i].y = b * rand() / RAND_MAX;
    }

    return points;
}

/*
 * Γράψτε μια συνάρτηση destructor για τον παραπάνω πίνακα
 */
void ptaDeletePoints(Point *pts) {
    /*
     * Αν είναι NULL, δεν υπάρχει κάτι να κάνουμε
     */
    if (pts == NULL) {
        return;
    }
    /*
     * Αλλιώς "αποδεσμεύουμε" τη μνήμη που είχαμε "δεσμεύσει"
     */
    free(pts);
}

/*
 * Γράψτε μια συνάρτηση που θα μετακινεί όλα τα σημεία του πίνακα κατά dx , dy.
 */
void ptaTranslatePoints( Point *pts, int N, double dx, double dy ) {
    /*
     * Αν είναι NULL ή άδειος ο πίνακας, δεν υπάρχει κάτι να κάνουμε
     */
    if (pts == NULL || N < 1) {
        return;
    }

    /*
     * Αλλιώς για κάθε σημείο από αυτά
     */
    for (int i = 0; i < N; ++i) {
        /*
         * Αύξησε την κάθε συντεταγμένη κατά την αντίστοιχη μεταβολή
         */
        pts[i].x += dx;
        pts[i].y += dy;
    }
}

/*
 * Γράψτε μία συνάρτηση που θα επιστρέφει το πλήθος των σημείων
 * ενός τέτοιου πίνακα, τα οποία βρίσκονται μέσα σε έναν κύκλο
 * ακτίνας R και κέντρου x , y.
 */
int ptaCountPointsInCircle(Point *pts, int N, double x, double y, double R) {
    /*
     * Πάντα ελέγχω τα δεδομένα, αν όντως δόθηκε κάτι και αν ο κύκλος έχει νόημα
     */
    if (pts == NULL || N < 1 || R <= 0) {
        return -1;
    }

    /*
     * Ορίζω έναν μετρητή και υπολογίζω μία φορά το τετράγωνο της ακτίνας
     */
    int count = 0;
    double R2 = R*R;
    /*
     * Για κάθε σημείο στον πίνακα
     */
    for (int i = 0; i < N; ++i) {
        /*
         * Υπολογίζω την απόστασή του από το κέντρο του κύκλου κατά x και y
         */
        double dx = pts[i].x - x;
        double dy = pts[i].y - y;
        /*
         * Ελέγχω εάν είναι μέσα στον κύκλο, αλλά για ταχύτητα το κάνω με
         * ύψωση στο τετράγωνο.
         */
        if (dx*dx + dy*dy <= R2) {
            /*
             * Εάν το σημείο είναι εντός του κύκλου, το προσμετρώ
             */
            count++;
        }
    }
    /*
     * Επιστρέφω το πλήθος των σημείων
     */
    return count;
}
/*
 * Γράψτε μια συνάρτηση που θα επιστρέφει έναν νέο (υπό)πίνακα με τέτοια σημεία,
 * τα οποία βρίσκονται μέσα σε έναν κύκλο ακτίνας R και κέντρου x , y για έναν
 * άλλο τέτοιο δεδομένο πίνακα σημείων.
 */
Point *ptaPointsInCircle( Point *pts, int N, double x, double y, double R ) {
    /*
     * Εδώ θα έπρεπε να γίνει έλεγχος των δεδομένων, αλλά η ptaCountPointsInCircle
     * που χρειάζεται αμέσως μετά, τους κάνει αυτή, οπότε δεν χρειάζεται να γίνουν
     * δύο φορές.
     */
    int countInCircle = ptaCountPointsInCircle(pts,N,x,y,R);
    /*
     * Αν κάτι δεν πήγε καλά (-1) ή εάν δεν χρειάζεται να κάνω κάτι (0), τότε
     * τερματίζω την εκτέλεση.
     */
    if (countInCircle < 1) {
        return NULL;
    }

    /*
     * Δημιουργώ έναν πίνακα που να χωράει όσα σημεία χρειάζονται, και για
     * ευκολία το κάνω με τον υπάρχοντα constructor.
     */
    Point *inCircle = ptaCreateRandomInBox(0,0,countInCircle);
    /*
     * Εάν δεν μπορεί να δημιουργθεί ο πίνακας που θα κρατήσει το αποτέλεσμα
     * τότε δεν μπορούμε να κάνουμε κάτι και τερματίζουμε.
     */
    if (inCircle == NULL) {
        return NULL;
    }

    /*
     * Ορίζω το τετράγωνο της ακτίνας και έναν μετρητή ο οποίος υποδεικνύει
     * την επόμενη θέση στον πίνακα του αποτελέσματος που θα πρέπει να μπει
     * το επόμενο σημείο που θα βρεθεί εντός κύκλου.
     */
    int c=0;
    double R2 = R*R;
    /*
     * Για κάθε σημείο στον πίνακα
     */
    for (int i = 0; i < N; ++i) {
        /*
         * Υπολογίζω την απόστασή του από το κέντρο του κύκλου κατά x και y
         */
        double dx = pts[i].x - x;
        double dy = pts[i].y - y;
        /*
         * Ελέγχω εάν είναι μέσα στον κύκλο, αλλά για ταχύτητα το κάνω με
         * ύψωση στο τετράγωνο.
         */
        if (dx*dx + dy*dy <= R2) {
            /*
             * Εάν το σημείο είναι εντός του κύκλου, το αντιγράφω στην
             * επόμενη διαθέσιμη θέση του αποτελέσματατος και αυξάνω
             * και αυτή τη θέση.
             */
            inCircle[c].x = pts[i].x;
            inCircle[c].y = pts[i].y;
            c++;
        }
    }

    /*
     * Επιστρέφω τον νέο πίνακα με τα εντός κύκλου σημεία
     */
    return inCircle;
}

/*
 * Με βάση τα παραπάνω φτιάχνουμε ένα απλό παράδειγμα το
 * οποίο υπολογίζει το π, προσεγγιστικά.
 *
 * Ορίζω μία σταθερά για το πλήθος των σημείων που θα
 * χρησιμοποιηθούν.
 */
#define SZ 10000000
int main() {
    /*
     * Προετοιμασία συστήματος/γεννήτριας τυχαίων αριθμών
     */
    srand(time(0));
    rand();

    /*
     * Δημιουργώ στο τετράγωνο 1x1 SZ σημεία
     */
    Point *box = ptaCreateRandomInBox(1,1,SZ);
    /*
     * Μετράω πόσα από αυτά είναι μέσα στον μοναδιαίο κύκλο.
     * Πρέπει να είναι κάτι λιγότερα από το 1/4.
     */
    int pic = ptaCountPointsInCircle(box, SZ, 0,0,1);

    /*
     * Το πηλίκο των εντός κύκλου (x4 επειδή το τετράγωνο
     * καλύπτει μόνο το ένα τεταρτημόριο) με τα συνολικά,
     * δίνει την αναλογία των επιφανειών, η οποία εάν την
     * υπολογίσω αναλυτικά είναι π (3,1415...) αλλάζοντας
     * το πλήθος των σημείων προσεγγίζω καλύτερα το π.
     *
     * ΠΡΟΣΟΧΗ! Επειδή έχουμε χρησιμοποιήσει INT τύπο
     * δεδομένων, δεν μπορούμε να ορίσουμε το SZ περισσότερο
     * από τη μέγιστη τιμή που μπορεί να παραστήσει ένας
     * integer.
     */
    printf("%lf\n", 4.0*pic/SZ);
    return 0;
}